import typing as t

from kit_up.core.clauses import base as bcls
from kit_up.core.clauses import predicates as dfs
from kit_up.core.domains import base
from kit_up.core.mixins import crud
from kit_up.core.repositories import crud as repos


class CrudDomain(
    base.AbstractDomain,
    crud.AbstractSemiCrud,
    repos.RequiresCrudRepositoryMixin,
):
    def create(self, input, *inputs) -> tuple:
        data = self._repository.create(input, *inputs)
        return tuple(model.as_view() for model in data)

    def pick(self, identity, *clauses: bcls.BaseClause):
        data = self._repository.pick(identity)
        return data.as_view()

    def filter(self, *clauses: bcls.BaseClause) -> t.Iterable:
        data = self._repository.filter(*clauses)
        return tuple(item.as_view() for item in data)

    def update(self, identity, updates) -> t.Any:
        model = self._repository.pick(identity)
        updated = model.clone(**updates)
        self._repository.update(updated)
        return updated.as_view()

    def _erase(self, *filters: dfs.AbstractPredicateClause):
        return self._repository.erase(*filters)

    def destroy(self, identity) -> None:
        return self._repository.destroy(identity)


class AsyncCrudDomain(
    base.AbstractAsyncDomain,
    crud.AbstractAsyncSemiCrud,
    repos.RequiresAsyncCrudRepositoryMixin,
):
    async def create(self, input, *inputs) -> tuple:
        data = await self._repository.create(input, *inputs)
        return tuple(model.as_view() for model in data)

    async def pick(self, identity, *clauses: bcls.BaseClause):
        data = await self._repository.pick(identity)
        return data.as_view()

    async def filter(self, *clauses: bcls.BaseClause) -> t.Iterable:
        data = await self._repository.filter(*clauses)
        return tuple(item.as_view() for item in data)

    async def update(self, identity, updates) -> t.Any:
        model = self._repository.pick(identity)
        updated = model.clone(**updates)
        await self._repository.update(updated)
        return updated.as_view()

    async def _erase(self, *filters: dfs.AbstractPredicateClause):
        return self._repository.erase(*filters)

    async def destroy(self, identity) -> None:
        return await self._repository.destroy(identity)
